#############################################################################
#
#      RECEPTOR RESPONSE CALCULATOR
#
# how to run the script
# note: all input files must have the same resolution
#
# example = ColourFilter() # initialize the object
#
# example.ReadInReceptors(receptor_path) # load the receptor spectral sensitivities from receptor_path, any number of receptor types is allowed
#
# example.ReadInIllumination(illumination_path) # load the spectrum of the illumination from illumination_path
#
# example.ReadInBackground(background_path) # load the reflectance spectrum of the background that the receptors are assumed to be adapted to from background_path
#
# example.ReadInStimulus(stimulus_path, stimulus_type,lambda_values_path,eyesize)
### load the stimulus data from stimulus_path
### possible stimulus_types:
###'single_spectrum_csv'- one csv file containing the stimulus lambda and reflectance values
###'single_spectrum_Fred_csv': # one csv file containing the stimulus lambda and reflectance values, with the ID in the first column (as it is from FReD)
###'hyperspectral_tiff': # multiple tiff files containing reflectance values measured with filters at those wavelengths
                         # the stimulus_path is the path to the first tiff of a set; names must be {name}{sequence number}.tif eg sample1.tif
### lambda_values_path: when using hyperspectral_tiff as stimulus type, give the path to a csv file containing the lambda values
### eyesize: when using hyperspectral_tiff as stimulus type, this optional parameter reduces the resolution of the image, where eyesize is the new number of rows
#
#
# example.GetAdaptation() # calculate the adaptation values for the receptors (necessary step)
#
# example.GetExcitation() # calculate the receptor responses
#
# example.ExportQuantumCatch2CSV() # save the calculated quantum catches to a csv file to the same library where the stimulus data is
#
# example.ExportExcitation2CSV() # save the calculated receptor responses to a csv file to the same library where the stimulus data is
#
# example.ExportExcitation2Png() # generate and save a heat map of the calculated receptor responses to the same library where the stimulus data is
#
# Please cite the following article when using this script:
# Vasas V, Hanley D, Kevan P, Chittka L (2017) Multispectral images of flowers reveal the adaptive significance of using long-wavelength-sensitive receptors for edge detection in bees. 2017, J Comp Physiol A doi:10.1007/s00359-017-1156-x
# For questions, contact Vera Vasas at v.vasas@qmul.ac.uk.


#############################################################################

import numpy as np
import scipy.optimize as scimin
from PIL import Image

from matplotlib import pyplot as plt


class Spectrum:
    # class for spectral data
    
    def __init__(self, lambda_values, y_values, lambda_min, lambda_max, lambda_step):
        
               
        self.lambda_values = np.arange(lambda_min, lambda_max+lambda_step,lambda_step) # wavelength range is the same for all the data, derived from receptor data
        self.y_values=np.empty((np.shape(self.lambda_values)[0],0))
        
        y_values_internal = np.reshape(y_values,(np.shape(y_values)[0],-1)) # convert the list to 2 dimensional array for furher calculations
        
        for column in y_values_internal.T:
                        
            self.y_values = np.column_stack((self.y_values,column))  # spectral y values are in a 2D array
        
            
class Picture:
    # class for hyperspectral data
    
    def __init__(self, lambda_values, y_values_picture, lambda_min, lambda_max, lambda_step):  
        
        if lambda_values[0] > lambda_min or lambda_values[-1] < lambda_max:  # the script still runs if spectral data is incomplete, the user must use sensible input ranges
            print "Stimulus reflectance data is incomplete for the range specified by receptors. Stimulus reflectance range:", lambda_values[0], lambda_values[-1], "Receptor range: ", lambda_min, lambda_max
                   
        self.lambda_values=np.arange(lambda_min, lambda_max+lambda_step,lambda_step) # wavelength range is the same for all the data, derived from receptor data
                
        self.y_values=[]
        
        for pixel in y_values_picture:  # y_values_picture list of spectral data for each pixel
            self.y_values.append(Spectrum(lambda_values,pixel,lambda_min, lambda_max, lambda_step).y_values)
      
class ColourFilter:
    # main; class for reading in and formatting input data, doing calculations on it, saving the output
    
    def __init__(self):  # do nothing else when a ColourFilter object is initialized
        
        pass
    
    def ReadInReceptors(self,receptor_path):   # read in the receptor sensitivity values from a csv file and extract the wavelength for which the all calculations will be done
        
       with open(receptor_path,'rb') as receptors_file:     # read in the names of the receptors
           receptor_header=receptors_file.readline().rstrip()
       self.receptor_names=receptor_header.split(',')[1:]
       
       receptor_data=np.genfromtxt(receptor_path,delimiter=',',skip_header=1)  # read in the receptor data as numpy array

       self.lambda_min=receptor_data[0,0]   # lambda values for all calculations are taken from the receptors' range
       self.lambda_max=receptor_data[-1,0]
       self.lambda_step=(float(self.lambda_max)-float(self.lambda_min))/(float(receptor_data.shape[0])-1.)
             
       self.receptor_sensitivity=Spectrum(receptor_data[:,0], receptor_data[:,1:], self.lambda_min, self.lambda_max, self.lambda_step)  # store receptor sensitivity values as a Spectrum object
       
       
    def ReadInIllumination(self,illumination_path):   # read in the illumination data
        
        illumination_data=np.genfromtxt(illumination_path,delimiter=',',skip_header=1)
        self.illumination=Spectrum(illumination_data[:,0], illumination_data[:,1], self.lambda_min, self.lambda_max, self.lambda_step)  # store the illumination data as a Spectrum object
               
        if illumination_data[0,0] > self.lambda_min or illumination_data[-1,0] < self.lambda_max:  # the script still runs if spectral data is incomplete,, the user must use sensible input ranges
            print "Illumination data is incomplete for the range specified by receptors. Illumination range:", illumination_data[0,0], illumination_data[-1,0], "Receptor range: ", self.lambda_min, self.lambda_max
    
    def ReadInBackground(self,background_path):   # read in the background reflectance to which the eye is adapted to; don't use zeros, it will end up badly
        
        background_data=np.genfromtxt(background_path,delimiter=',',skip_header=1)
        self.background_reflectance=Spectrum(background_data[:,0], background_data[:,1], self.lambda_min, self.lambda_max, self.lambda_step)  # store the background illumination values as a Spectrum object
         
        if background_data[0,0] > self.lambda_min or background_data[-1,0] < self.lambda_max:  # the script still runs if spectral data is incomplete,, the user must use sensible input ranges
            print "Background illumination data is incomplete for the range specified by receptors. Background illumination range:", background_data[0,0], background_data[-1,0], "Receptor range: ", self.lambda_min, self.lambda_max
        
    def ReadInStimulus(self,stimulus_path, stimulus_data_type, lambda_values_path=None, newsize_height=None): # read in the stimulus reflectance values; if resizing needed, give the required height
        
        self.stimulus_name=stimulus_path.split('.')[0]  # save the name of the file; cut off the extension
        
        if stimulus_data_type == 'single_spectrum_csv': # one csv file having the stimulus lambda and reflectance values
            
            stimulus_data=np.genfromtxt(stimulus_path,delimiter=',',skip_header=1)
            stimulus_y_values=[]
            stimulus_y_values.append(stimulus_data[:,1])
            
            self.picture_shape=(1,1)
            
            self.stimulus_reflectance=Picture(stimulus_data[:,0],stimulus_y_values,self.lambda_min, self.lambda_max, self.lambda_step)
        
        if stimulus_data_type == 'single_spectrum_Fred_csv': # one csv file having the stimulus lambda and reflectance values, with the ID in the first column (as it is from FReD)
            
            stimulus_data=np.genfromtxt(stimulus_path,delimiter=',',skip_header=1)
            stimulus_y_values=[]
            stimulus_y_values.append(stimulus_data[:,2])
            
            self.picture_shape=(1,1)
            
            self.stimulus_reflectance=Picture(stimulus_data[:,1],stimulus_y_values,self.lambda_min, self.lambda_max, self.lambda_step)
            
        if stimulus_data_type == 'hyperspectral_tiff': # one csv file having the lambda values, plus multiple tiff files containing reflectance values measured with filters at those wavelengths
                                                       # the stimulus_path is the first tiff of a set; names must be {name}{sequence number}.tif eg sample1.tif
            
            lambda_values_data=np.genfromtxt(lambda_values_path,delimiter=',',skip_header=1)
            number_location=stimulus_path.find('.')-1   #let's find out where the numbers are in the name
            
            stim_original_size= Image.open(stimulus_path).size   # resize the image if it was requested
            if newsize_height!=None:
                newsize=(int(float(newsize_height)/float(stim_original_size[1])*float(stim_original_size[0])),newsize_height)
                        
            if newsize_height!=None:    # start building the 3D image with the tiff given as input; resize if parameter is given                       
                threeDimage=np.array(Image.open(stimulus_path).resize(newsize, Image.ANTIALIAS))/255. 
            else:   
                threeDimage=np.array(Image.open(stimulus_path))/255.
                
            for i in range(1,np.size(lambda_values_data)): # read in the remaining layers one by one
                  
                layer_number=i+int(stimulus_path[number_location]) # number of the file that is due for reading in
                
                if np.size(lambda_values_data) < 10:   # sequence numbering is one-digit
                    layer_file_name=stimulus_path[0:number_location]+str(layer_number)+'.tif'  # generate the file name
                
                if 10 <= np.size(lambda_values_data) < 100:   # sequence numbering is two-digit # THIS NEEDS TO BE TESTED STILL
                    if layer_number < 10:
                        layer_file_name=stimulus_path[0:number_location-1]+'0'+str(layer_number)+'.tif'  # generate the file name
                    if 10 <=layer_number < 100:
                        layer_file_name=stimulus_path[0:number_location-1]+str(layer_number)+'.tif'  # generate the file name
                
                if newsize_height!=None: # read in the current layer as an np array, resize if parameter is given
                    layer=np.array(Image.open(layer_file_name).resize(newsize, Image.ANTIALIAS))/255. 
                else:
                    layer=np.array(Image.open(layer_file_name))/255. 
                                          
                threeDimage = np.dstack((threeDimage,layer))  # add the current layer to the 3D array
                                   
            stimulus_y_values=np.reshape(threeDimage,(-1,np.shape(threeDimage)[2])).tolist()  # convert the numpy array to a list pixels  
                        
            self.picture_shape=(np.shape(threeDimage)[0],np.shape(threeDimage)[1]) # tuplet about row and column number if it is a hyperspectral image
            
            self.stimulus_reflectance=Picture(lambda_values_data,stimulus_y_values,self.lambda_min, self.lambda_max, self.lambda_step)
            
            
    def GetAdaptation(self):
        # calculate the receptor's adaptation state when adapted to a background
        self.receptor_adaptation = 1 / sum(self.receptor_sensitivity.y_values * self.illumination.y_values * self.background_reflectance.y_values)
      
        
    def GetExcitation(self):
        # calculate the receptor excitation values
        self.quantum_catch_list=list(self.receptor_adaptation * sum(self.receptor_sensitivity.y_values * self.illumination.y_values * pixel) for pixel in self.stimulus_reflectance.y_values)
        self.excitation_list=list(pixel/(1+pixel) for pixel in self.quantum_catch_list)
        
    
    def QuantumCatchList2threeDarray(self):
        # convert the list of pixels to a three-D array for further handling where x=row y=column in the picture and z=receptor type
        self.quantum_catch_array= np.reshape(np.array(self.quantum_catch_list),(self.picture_shape[0],self.picture_shape[1],np.shape(self.receptor_sensitivity.y_values)[1]))
    
    def ExcitationList2threeDarray(self):
        # convert the list of pixels to a three-D array for further handling where x=row y=column in the picture and z=receptor type
        self.excitation_array= np.reshape(np.array(self.excitation_list),(self.picture_shape[0],self.picture_shape[1],np.shape(self.receptor_sensitivity.y_values)[1]))
    
    def ExportQuantumCatch2CSV(self):
        # save the quantum catch values as csv files, separate for each receptor
        
        if hasattr(self,'quantum_catch_array')==False:   # convert the quantum catch list to quantum catch array if it has not yet been done
            self.QuantumCatchList2threeDarray()
        
        for receptor in self.receptor_names:    # save to separate files 
            receptor_ID=self.receptor_names.index(receptor)
            receptor_file_name=self.stimulus_name+'_'+receptor+'_quantum_catch.csv'  # generate the name for saving from the filename of the stimulus and from the name of the receptor
            np.savetxt(receptor_file_name, self.quantum_catch_array[:,:,receptor_ID], delimiter=",")    
                            
    def ExportExcitation2CSV(self):
        # save the excitation values as csv files, separate for each receptor
        
        if hasattr(self,'excitation_array')==False:   # convert the excitation list to excitation array if it has nt yet been done
            self.ExcitationList2threeDarray()
        
        for receptor in self.receptor_names:    # save to separate files 
            receptor_ID=self.receptor_names.index(receptor)
            receptor_file_name=self.stimulus_name+'_'+receptor+'.csv'  # generate the name for saving from the filename of the stimulus and from the name of the receptor
            np.savetxt(receptor_file_name, self.excitation_array[:,:,receptor_ID], delimiter=",")
        
    def ExportExcitation2Png(self):
        # generate and save PNG files for receptors
        
        if hasattr(self,'excitation_array')==False:   # convert the excitation list to excitation array if it has not yet been done
            self.ExcitationList2threeDarray()
            
        for receptor in self.receptor_names:    # save to separate files 
            
            receptor_ID=self.receptor_names.index(receptor)
                                               
            layer=self.excitation_array[:,:,receptor_ID]
            
            receptor_file_name_heatmap=self.stimulus_name+'_'+receptor+'_hot.png'  # generate the name for saving the heatmap from the filename of the stimulus and from the name of the receptor
            
            plt.close()
            fig, ax = plt.subplots(figsize=(8, float(self.picture_shape[0])/float(self.picture_shape[1])*8. ))
            h=ax.pcolor(np.flipud(layer), cmap=plt.cm.hot, vmin=0., vmax=1.)
            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            h.axes.set_xlim(0, np.shape(layer)[1])
            h.axes.set_ylim(0, np.shape(layer)[0])
            fig.savefig(receptor_file_name_heatmap,bbox_inches='tight') #
            plt.close()  